// **********************************************************************
//
// Copyright (c) 2001 MT Tools.
//
// All Rights Reserved
//
// MT_DORB is based in part on the product DORB,
// written by Shadrin Victor
//
// See Readme.txt for contact information
//
// **********************************************************************
unit imr;

interface

uses
  orb,orb_int,imr_int,boa_int,types,req_int,env_int,classes,
  code_int,stdstat,ir_int;

type

  TObjectInfo = class(TInterfacedObject,IObjectInfo)
  private
    FRepoid: string;
    FTag : string;
  protected
    function repoid(): string; overload;
    function tag(): string; overload;
    procedure repoid(val: string); overload;
    procedure tag(val: string); overload;
  public
    constructor Create(repoid: string = ''; tag : string = '');
  end;

  TImplementationBase = class(TORBObject,IImplementationBase)
  protected
    FBOA : IBOA;
    //procedure create_ref(const id: string;const ImplBase: IUnknown;const repoid: string);
    procedure create_ref(const id: string;const idef: IInterfaceDef; const implDef : IImplementationDef;const repoid: string);
    function find_impl(const repid: string; const name: string): IImplementationDef;
    function impl_name(const name: string): string;
    function make_request(const req: IORBRequest;const obj: IORBObject; mid: ORBMessageID;
       const oa: IObjectAdaptor; pr: IPrincipal):IServerRequestBase; virtual;
    procedure doinvoke(const req: IServerRequestBase; const env: IEnvironment); virtual;
  public
    constructor Create();
  end;

  TStaticImplementation = class(TImplementationBase,IStaticImplementation)
  protected
    procedure doinvoke(const req: IServerRequestBase; const env: IEnvironment); override;
    procedure invoke(const req: IStaticServerRequest; const env: IEnvironment); virtual;
    function make_request(const req: IORBRequest;const obj: IORBObject; id: ORBMessageID;
       const oa: IObjectAdaptor; pr: IPrincipal):IServerRequestBase; override;
  end;

  TStaticMethodDispatcher = class(TStaticImplementation,IStaticMethodDispatcher)
  private
    FDispatchers : IInterfaceList;
  protected
    procedure register_dispatcher(const disp: IStaticInterfaceDispatcher);
    procedure invoke(const req: IStaticServerRequest; const env: IEnvironment); override;
  public
    constructor Create();
  end;

  TImplementationDef_stub = class(TImplementationBase,IImplementationDef)
  protected
    function name(): string;
  public
    class function narrow(const obj: IORBObject): IImplementationDef;
  end;

  TImplementationDef_skel = class(TStaticMethodDispatcher,IImplementationDef,IStaticMethodDispatcher)
  protected
    function name(): string; virtual;
  end;

  TImplRepository_skel = class(TImplementationBase,IImplRepository)
  protected
    function narrow_helper(const repoid: string): Pointer;  override;
    function find_by_name(name: string): IInterfaceList; virtual;
    function _create(mode : TActivationMode;const objs: IInterfaceList;
      name: string; command: string): IImplementationDef; virtual;
  public
    constructor Create();
    class function narrow(const obj: IORBObject): IImplRepository;
  end;

  TImplRepository_stub = class(TImplementationBase,IImplRepository)
  protected
    function find_by_name(name: string): IInterfaceList;
    function _create(mode : TActivationMode;const objs: IInterfaceList;
      name: string;command: string): IImplementationDef;
  end;

  TMarshaller_seq_CORBA_ImplementationDef = class(TAbstractStat,IStaticTypeInfo)
    function demarshal(dec: IDecoder;const addr : Pointer): Boolean;
    procedure marshal(enc: IEncoder; const addr : Pointer);
    function typecode(): ITypeCode; override;
    procedure _create(var addr: Pointer); override;
    procedure _free(var addr: Pointer);
  end;

  TMarshaller_CORBA_ImplementationDef = class(TAbstractStat,IStaticTypeInfo)
    function demarshal(dec: IDecoder;const addr : Pointer): Boolean;
    procedure marshal(enc: IEncoder; const addr : Pointer);
    function typecode(): ITypeCode; override;
    procedure _create(var addr: Pointer); override;
    procedure _free(var addr: Pointer);
  end;

  TMarshaller_CORBA_ImplementationDef_ActivationMode = class(TAbstractStat,IStaticTypeInfo)
    function demarshal(dec: IDecoder;const addr : Pointer): Boolean;
    procedure marshal(enc: IEncoder; const addr : Pointer);
    function typecode(): ITypeCode; override;
    procedure _create(var addr: Pointer); override;
    procedure _free(var addr: Pointer);
  end;

  TMarshaller_seq_CORBA_ImplementationDef_ObjectInfo = class(TAbstractStat,IStaticTypeInfo)
    function demarshal(dec: IDecoder;const addr : Pointer): Boolean;
    procedure marshal(enc: IEncoder; const addr : Pointer);
    function typecode(): ITypeCode; override;
    procedure _create(var addr: Pointer); override;
    procedure _free(var addr: Pointer);
  end;

  TMarshaller_CORBA_ImplementationDef_ObjectInfo = class(TAbstractStat,IStaticTypeInfo)
    function demarshal(dec: IDecoder;const addr : Pointer): Boolean;
    procedure marshal(enc: IEncoder; const addr : Pointer);
    function typecode(): ITypeCode; override;
    procedure _create(var addr: Pointer); override;
    procedure _free(var addr: Pointer);
  end;


var
  _marshaller_seq_CORBA_ImplementationDef : IStaticTypeInfo;
  _marshaller_CORBA_ImplementationDef : IStaticTypeInfo;
  _marshaller_CORBA_ImplementationDef_ActivationMode : IStaticTypeInfo;
  _marshaller_seq_CORBA_ImplementationDef_ObjectInfo : IStaticTypeInfo;
  _marshaller_CORBA_ImplementationDef_ObjectInfo : IStaticTypeInfo;

implementation

uses
  sysutils,static,throw,exceptions;

//***************************************************************************
//  TImplementationBase
//***************************************************************************
constructor TImplementationBase.Create();
begin
  inherited Create();
  FBOA := FORB.BOA_Init as IBOA;
end;

procedure TImplementationBase.doinvoke(const req: IServerRequestBase; const env: IEnvironment);
begin
  assert(false,'must be implemented in descendants');
end;

function TImplementationBase.make_request(const req: IORBRequest;const obj: IORBObject; mid: ORBMessageID;
       const oa: IObjectAdaptor; pr: IPrincipal):IServerRequestBase;
begin
  assert(false,'must be implemented in descendants');
  result := nil;
end;

procedure TImplementationBase.create_ref(const id: string;const idef: IInterfaceDef; const implDef : IImplementationDef;const repoid: string);
var
  obj : IORBObject;
begin
  obj := FBOA.createrec(id,idef,impldef,IImplementationBase(self),repoid);
  Assign(obj);
end;

function TImplementationBase.find_impl(const repid: string; const name: string): IImplementationDef;
var
  obj: IORBObject;
  local_imr,_imr : IImplRepository;
  _name : string;
  lst : IInterfaceList;
  impl : IImplementationDef;
  oi: IObjectInfo;
begin
  result := nil;
  _name := impl_name(name);
  obj := FORB.resolve_initial_reference('ImplementationRepository');
  assert(obj <> nil);
  _imr := TImplRepository_skel.narrow(obj);
  assert(_imr<> nil);
  obj := FORB.resolve_initial_reference('LocalImplementationRepository');
  assert(obj <> nil);
  local_imr := TImplRepository_skel.narrow(obj);
  lst := _imr.find_by_name(_name);
  if lst.Count = 0 then
    lst := local_imr.find_by_name(_name);
  if lst.Count > 0 then
    begin
      assert(false,'not implemented');
    end
  else
    begin
      oi := TObjectInfo.Create(repid);
      lst.Add(oi);
      impl := local_imr._create(ActivatePersistent,lst,name,'');
      result := impl;
    end;
end;

function TImplementationBase.impl_name(const name: string): string;
begin
  result := FBOA.impl_name
end;

//************************************************************************
//  TImplRepository
//************************************************************************
function TImplRepository_skel.narrow_helper(const repoid: string): Pointer;
begin
  result := nil;
  if strcomp(PChar(repoid),'IDL:omg.org/CORBA/ImplRepository:1.0') = 0 then
    result := Pointer(IImplRepository(self));
end;

function TImplRepository_skel.find_by_name(name: string): IInterfaceList;
begin
  result := nil;
  // must be implemented in descendant classes
end;

class function TImplRepository_skel.narrow(const obj: IORBObject): IImplRepository;
begin
  result := nil;
  if obj = nil then exit;
  if obj.narrow_helper('IDL:omg.org/CORBA/ImplRepository:1.0') <> nil then
    begin
      result := IImplRepository(obj);
      exit;
    end
  else
    assert(false,'not implemented');
end;

function TImplRepository_skel._create(mode : TActivationMode;const objs: IInterfaceList;
      name: string; command: string): IImplementationDef;
begin
  assert(false,'must be implemented in descendant classes');
end;

//************************************************************************
// TImplRepository_skel
//************************************************************************
constructor TImplRepository_skel.Create();
begin
  inherited Create();
  create_ref('',nil,nil,'omg.org/CORBA/ImplRepository:1.0');
end;

//*********************************************************************
//  TImplRepository_stub
//*********************************************************************
function TImplRepository_stub.find_by_name(
  name: string): IInterfaceList;
var
  _name,_res : IStaticAny;
  req : IStaticRequest;
begin
  _name := TStaticAny.Create(stat_String,PString(@name));
  _res := TStaticAny.Create(_marshaller_seq_CORBA_ImplementationDef);
  req := TStaticRequest.Create(self,'find_by_name');
  req.add_in_arg(_name);
  req.set_result(_res);
  req.invoke;
  dorb_static_throw(req);
  result := IInterfaceList(_res.value^);
end;

function TImplRepository_stub._create(mode: TActivationMode;
  const objs: IInterfaceList; name: string; command: string): IImplementationDef;
var
  _mode, _objs, _name, _command: IStaticAny;
begin
  _mode := TStaticAny.Create(_marshaller_CORBA_ImplementationDef_ActivationMode,@mode);
  _objs := TStaticAny.Create(_marshaller_seq_CORBA_ImplementationDef_ObjectInfo,@objs);
  _name := TStaticAny.Create(stat_String,PString(@name));
  _command := TStaticAny.Create(stat_String,PString(@command));

end;

//*********************************************************************
//  TImplementationDef_stub
//*********************************************************************
class function TImplementationDef_stub.narrow(
  const obj: IORBObject): IImplementationDef;
begin
  result := nil;
  assert(false,'not implemented');
end;

function TImplementationDef_stub.name(): string;
var
  _res : IStaticAny;
  req : IStaticRequest;
begin
  _res := TStaticAny.Create(stat_String,PString(@result));
  req := TStaticRequest.Create(self,'_get_name');
  req.set_result(_res);
  req.invoke;
  dorb_static_throw(req);
end;

//*********************************************************************
//  TObjectInfo
//*********************************************************************
constructor TObjectInfo.Create(repoid, tag: string);
begin
  FRepoid := repoid;
  FTag := tag;
end;

procedure TObjectInfo.repoid(val: string);
begin
  FRepoid := val;
end;

procedure TObjectInfo.tag(val: string);
begin
  FTag := val;
end;

function TObjectInfo.repoid: string;
begin
  result := FRepoid;
end;

function TObjectInfo.tag: string;
begin
  result := FTag;
end;

//*************************************************
//  TStaticImplementation
//*************************************************
procedure TStaticImplementation.doinvoke(const req: IServerRequestBase;
  const env: IEnvironment);
begin
  invoke(IStaticServerRequest(req),env);
end;

procedure TStaticImplementation.invoke(const req: IStaticServerRequest;
  const env: IEnvironment);
begin
  assert(false,'must be implemented in descendant classes');
end;

function TStaticImplementation.make_request(const req: IORBRequest;const obj: IORBObject; id: ORBMessageID;
       const oa: IObjectAdaptor; pr: IPrincipal):IServerRequestBase;
begin
  result := TStaticServerRequest.Create(req,obj,id,oa,pr)
end;

//*************************************************
//  TStaticMethodDispatcher
//*************************************************
constructor TStaticMethodDispatcher.Create;
begin
  inherited Create();
  FDispatchers := TInterfaceList.Create;
end;

procedure TStaticMethodDispatcher.invoke(const req: IStaticServerRequest;
  const env: IEnvironment);
var
  i : integer;
begin
  try
    for i := 0 to Pred(FDispatchers.Count) do
      if IStaticInterfaceDispatcher(FDispatchers[i])._dispatch(req,env) then
        exit;
  except
    on e : SystemException do
      begin
        req.set_exception(SystemException.Create(e));
        req.write_results();
      end
    else
      begin
        req.set_exception(SystemException.Create_type(st_UNKNOWN,0,COMPLETED_NO));
        req.write_results();
      end;
  end;
  req.set_exception(SystemException.Create_type(st_BAD_OPERATION));
  req.write_results;
end;

procedure TStaticMethodDispatcher.register_dispatcher(
  const disp: IStaticInterfaceDispatcher);
begin
  FDispatchers.Add(disp);
end;

//*************************************************
//  TImplementationDef_skel
//*************************************************
function TImplementationDef_skel.name: string;
begin
  assert(false,'not imlplemented');
end;

//*********************************************************************
//  TMarshaller__seq_CORBA_ImplementationDef
//*********************************************************************
procedure TMarshaller_seq_CORBA_ImplementationDef._create(var addr: Pointer);
begin
  IInterfaceList(addr^) := TInterfaceList.Create;
end;

procedure TMarshaller_seq_CORBA_ImplementationDef._free(var addr: Pointer);
begin
  IInterfaceList(addr^) := nil;
end;

function TMarshaller_seq_CORBA_ImplementationDef.demarshal(dec: IDecoder;
  const addr: Pointer): Boolean;
var
  v : IInterfaceList;
  n : integer;
  len: _ulong;
  temp : IImplementationDef;
begin
  result := false;
  if Pointer(addr^) = nil then
    IInterfaceList(addr^) := TInterfaceList.Create;
  if not dec.seq_begin(len) then exit;
  v := IInterfaceList(addr^);
  v.Capacity := len;
  for n := 0 to len - 1 do
    begin
      temp := TImplementationDef_stub.Create;
      if not _marshaller_CORBA_ImplementationDef.demarshal(dec,Pointer(@temp)) then exit;
      v.Add(temp);
    end;
  if not dec.seq_end then exit;
  result := true;
end;

procedure TMarshaller_seq_CORBA_ImplementationDef.marshal(enc: IEncoder;
  const addr: Pointer);
var
  lst: IInterfaceList;
  n : integer;
  imd : IImplementationDef;
begin
  lst := IInterfaceList(addr^);
  enc.seq_begin(lst.Count);
  for n := 0 to Pred(Lst.Count) do
    begin
       imd := IImplementationDef(lst[n]);
      _marshaller_CORBA_ImplementationDef.marshal(enc,Pointer(@imd));
    end;
  enc.seq_end;
end;

function TMarshaller_seq_CORBA_ImplementationDef.typecode: ITypeCode;
begin
  assert(false,'not implemented');
end;

//*********************************************************************
//  TMarshaller_CORBA_ImplementationDef
//*********************************************************************
procedure TMarshaller_CORBA_ImplementationDef._create(var addr: Pointer);
begin
  IImplementationDef(addr^) := TImplementationDef_stub.Create;
end;

procedure TMarshaller_CORBA_ImplementationDef._free(var addr: Pointer);
begin
  IImplementationDef(addr^) := TImplementationDef_stub.Create;
end;

function TMarshaller_CORBA_ImplementationDef.demarshal(dec: IDecoder;
  const addr: Pointer): Boolean;
var
  obj : IORBObject;
  imd : IImplementationDef;
begin
  result := false;
  if not stat_Object.demarshal(dec,Pointer(@obj)) then exit;
  imd := TImplementationDef_stub.narrow(obj);
  IORBObject(addr^) := IORBObject(imd);
  result := (obj = nil) or (imd <> nil);
end;

procedure TMarshaller_CORBA_ImplementationDef.marshal(enc: IEncoder;
  const addr: Pointer);
begin
  stat_Object.marshal(enc,addr);
end;

function TMarshaller_CORBA_ImplementationDef.typecode: ITypeCode;
begin
  assert(false,'not implemented');
end;

//*********************************************************************
//  TMarshaller_CORBA_ImplementationDef_ActivationMode
//*********************************************************************
procedure TMarshaller_CORBA_ImplementationDef_ActivationMode._create(var addr: Pointer);
begin
  addr := AllocMem(sizeof(TActivationMode));
end;

procedure TMarshaller_CORBA_ImplementationDef_ActivationMode._free(var addr: Pointer);
begin
  freemem(addr);
end;

function TMarshaller_CORBA_ImplementationDef_ActivationMode.demarshal(
  dec: IDecoder; const addr: Pointer): Boolean;
var
  ul: _ulong;
begin
  result := false;
  if not dec.get_ulong(ul) then exit;
  TActivationMode(addr^) := TActivationMode(ul);
end;

procedure TMarshaller_CORBA_ImplementationDef_ActivationMode.marshal(
  enc: IEncoder; const addr: Pointer);
begin
  enc.put_ulong(_ulong(addr^));
end;

function TMarshaller_CORBA_ImplementationDef_ActivationMode.typecode: ITypeCode;
begin
  assert(false,'not implemented');
end;

//*********************************************************************
//  TMarshaller_seq_CORBA_ImplementationDef_ObjectInfo
//*********************************************************************
procedure TMarshaller_seq_CORBA_ImplementationDef_ObjectInfo._create(var addr: Pointer);
begin
  IInterfaceList(addr^) := TInterfaceList.Create;
end;

procedure TMarshaller_seq_CORBA_ImplementationDef_ObjectInfo._free(var addr: Pointer);
begin
  IInterfaceList(addr^) := nil;
end;

function TMarshaller_seq_CORBA_ImplementationDef_ObjectInfo.demarshal(
  dec: IDecoder; const addr: Pointer): Boolean;
var
  v : IInterfaceList;
  n : integer;
  len: _ulong;
  temp : IImplementationDef;
begin
  result := false;
  if Pointer(addr^) = nil then
    IInterfaceList(addr^) := TInterfaceList.Create;
  if not dec.seq_begin(len) then exit;
  v := IInterfaceList(addr^);
  v.Capacity := len;
  for n := 0 to len - 1 do
    begin
      temp := TImplementationDef_stub.Create;
      if not _marshaller_CORBA_ImplementationDef_ObjectInfo.demarshal(dec,Pointer(@temp)) then exit;
      v.Add(temp);
    end;
  if not dec.seq_end then exit;
  result := true;
end;

procedure TMarshaller_seq_CORBA_ImplementationDef_ObjectInfo.marshal(
  enc: IEncoder; const addr: Pointer);
var
  lst: IInterfaceList;
  n : integer;
  imd : IImplementationDef;
begin
  lst := IInterfaceList(addr^);
  enc.seq_begin(lst.Count);
  for n := 0 to Pred(Lst.Count) do
    begin
       imd := IImplementationDef(lst[n]);
      _marshaller_CORBA_ImplementationDef_ObjectInfo.marshal(enc,Pointer(@imd));
    end;
  enc.seq_end;
end;

function TMarshaller_seq_CORBA_ImplementationDef_ObjectInfo.typecode: ITypeCode;
begin
  assert(false,'not implemented');
end;

//*********************************************************************
//  TMarshaller_CORBA_ImplementationDef_ObjectInfo
//*********************************************************************
procedure TMarshaller_CORBA_ImplementationDef_ObjectInfo._create(var addr: Pointer);
begin
  IObjectInfo(addr^) := TObjectInfo.Create();
end;

procedure TMarshaller_CORBA_ImplementationDef_ObjectInfo._free(var addr: Pointer);
begin
  IObjectInfo(addr^) := nil;
end;


function TMarshaller_CORBA_ImplementationDef_ObjectInfo.demarshal(
  dec: IDecoder; const addr: Pointer): Boolean;
var
  info: IObjectInfo;
  str: string;
begin
  result := false;
  if Pointer(addr^) = nil then
    IObjectInfo(addr^) := TObjectInfo.Create();
  info := IObjectInfo(Pointer(addr^));
  if not dec.get_string(str) then exit;
  info.repoid(str);
  if not dec.get_string(str) then exit;
  info.tag(str);
  result := true;
end;

procedure TMarshaller_CORBA_ImplementationDef_ObjectInfo.marshal(
  enc: IEncoder; const addr: Pointer);
var
  info: IObjectInfo;
begin
  info := IObjectInfo(Pointer(addr^));
  enc.put_string(info.repoid);
  enc.put_string(info.tag);
end;

function TMarshaller_CORBA_ImplementationDef_ObjectInfo.typecode: ITypeCode;
begin
  assert(false,'not implemented');
end;

initialization
  _marshaller_seq_CORBA_ImplementationDef := TMarshaller_seq_CORBA_ImplementationDef.Create;
  _marshaller_CORBA_ImplementationDef := TMarshaller_CORBA_ImplementationDef.Create;
  _marshaller_CORBA_ImplementationDef_ActivationMode := TMarshaller_CORBA_ImplementationDef_ActivationMode.Create;
  _marshaller_seq_CORBA_ImplementationDef_ObjectInfo := TMarshaller_seq_CORBA_ImplementationDef_ObjectInfo.Create;
  _marshaller_CORBA_ImplementationDef_ObjectInfo := TMarshaller_CORBA_ImplementationDef_ObjectInfo.Create;
finalization
  _marshaller_seq_CORBA_ImplementationDef := nil;
  _marshaller_CORBA_ImplementationDef := nil;
  _marshaller_CORBA_ImplementationDef_ActivationMode := nil;
  _marshaller_seq_CORBA_ImplementationDef_ObjectInfo := nil;
  _marshaller_CORBA_ImplementationDef_ObjectInfo := nil;
end.
