// **********************************************************************
//
// Copyright (c) 2001 MT Tools.
//
// All Rights Reserved
//
// MT_DORB is based in part on the product DORB,
// written by Shadrin Victor
//
// See Readme.txt for contact information
//
// **********************************************************************
unit ir_value;

interface

uses
  ir_int,orb_int,orb,classes,types,code_int,ir;

type

  TValueDef = class(TContainer,IValueDef,IContained,IIDLType)
  protected
    FTypeCode: ITypeCode;
    FCustom,FAbstract,FTruncatable : Boolean;
    FBaseValue: IValueDef;
    FAbstractBaseValues: IInterfaceList;
    FSupportedInt : IInterfaceList;
    FInitializers : InitializerSeq;
  protected
    function get_supported_interfaces: IInterfaceList;
    procedure set_supported_interfaces(const val: IInterfaceList);
    function initializers: InitializerSeq;
    procedure resize_initializers(len: integer);
    function get_base_value: IValueDef;
    procedure set_base_value(const val: IValueDef);
    function get_abstract_base_values: IInterfaceList;
    procedure set_abstract_base_values(const val: IInterfaceList);
    function is_abstract: Boolean;
    procedure set_is_abstract(val: Boolean);
    function is_custom: Boolean;
    procedure set_is_custom(val: Boolean);
    function is_truncatable: Boolean;
    procedure set_is_truncatable(val: Boolean);
    function create_value_member(id,name,version: string; mtype: IIDLType; acc: TVisibility): IValueMemberDef;
    function create_atribute(id,name,version: string; atype: IIDLType; amode: AttributeMode): IAttributeDef;
    function create_operation(id,name,version: string; oresult: IIDLType;
      omode: TOperationMode; const params: ParDescriptionSeq = nil; exceptions: ExceptionDefSeq = nil): IOperationDef;
    // IIDLType methods
    function _type(): ITypeCode;
    function narrow_helper(const repoid: string): Pointer; override;
    procedure check_overloaded(id,oname: string);
    function describe_value(): FullValueDescription;
  public
    constructor Create(const con: IContainer;id, name, version: string;
      is_custom,is_abstract: Boolean; base_value: IValueDef; is_truncatable: Boolean; abstract_base_values,
      supported_interfaces: IInterfaceList; initializers: PInitializerSeq);
    class function _narrow(const obj: IORBObject): IValueDef;
  end;

  TValueMemberDef = class(TContained,IValueMemberDef)
  protected
    FType : IIDLType;
    FAccess : TVisibility;
  protected
    function _type: ITypeCode;
    function get_type_def: IIDLType;
    procedure set_type_def(const val: IIDLType);
    function get_access: TVisibility;
    procedure set_access(val: TVisibility);
    function narrow_helper(const repoid: string): Pointer; override;
  public
    constructor Create(const con: IContainer;id,name,version: string; mtype: IIDLType; acc: TVisibility);
    class function _narrow(const obj: IORBObject): IValueMemberDef;
  end;


implementation

uses
  throw,any,tcode;


//*********************************************************************
//  TValueDef
//*********************************************************************
function TValueDef._type: ITypeCode;
var
  modifier : TValueModifier;
  members : ValueMemberSeq;
  i,j: integer;
  mem: IValueMemberDef;
  base_type: ITypeCode;
begin
  if FTypeCode = nil then
    begin
      modifier := VM_NONE;
      if FAbstract then
        modifier := VM_ABSTRACT;
      if FCustom then
        modifier := VM_CUSTOM;
      if FTruncatable then
        modifier := VM_TRUNCATABLE;
      j := 0;
      for i := 0 to Pred(FContained.Count) do
        begin
          mem := TValueMemberDef._narrow(FContained[i] as IORBObject);
          if mem = nil then continue;
          setlength(members,j+1);
          members[j].name := mem.name;
          members[j].id := mem.id;
          members[j].version := mem.version;
          members[j]._type := mem._type;
          inc(j);
        end;
      if FBaseValue <> nil then
        base_type := FBaseValue._type
      else
        base_type := nil;
      FTypeCode := TTypeCode.create_value_tc(FID,FName,modifier,base_type,members);
    end;
  result := FTypeCode;
end;

function TValueDef.create_atribute(id, name, version: string;
  atype: IIDLType; amode: AttributeMode): IAttributeDef;
begin
  check_overloaded(id,name);
  result := TAttributeDef.Create(self,id,name,version,atype,amode);
  FContained.Add(result as IContained);
end;

function TValueDef.create_operation(id,name,version: string; oresult: IIDLType;
      omode: TOperationMode; const params: ParDescriptionSeq; exceptions: ExceptionDefSeq): IOperationDef;
begin
  check_overloaded(id,name);
  result := TOperationDef.Create(self,id,name,version,oresult,omode,params,exceptions);
  FContained.Add(result as IContained);
end;

function TValueDef.create_value_member(id, name, version: string;
  mtype: IIDLType; acc: TVisibility): IValueMemberDef;
begin
  result := TValueMemberDef.Create(self,id,name,version,mtype,acc);
  FContained.add(result as IContained)
end;

function TValueDef.get_abstract_base_values: IInterfaceList;
begin
  result := FAbstractBaseValues;
end;

function TValueDef.get_base_value: IValueDef;
begin
  result := FBaseValue;
end;

function TValueDef.initializers: InitializerSeq;
begin
  result := FInitializers;
end;

procedure TValueDef.resize_initializers(len: integer);
begin
  setLength(FInitializers,len);
end;

function TValueDef.get_supported_interfaces: IInterfaceList;
begin
  result := FSupportedInt
end;

function TValueDef.is_abstract: Boolean;
begin
  result := FAbstract;
end;

function TValueDef.is_custom: Boolean;
begin
  result := FCustom;
end;

function TValueDef.is_truncatable: Boolean;
begin
  result := FTruncatable
end;

function TValueDef.narrow_helper(const repoid: string): Pointer;
begin
  if repoid = 'IDL:omg.org/CORBA/ValueDef:1.0' then
    result := Pointer(IValueDef(self))
  else if repoid = 'IDL:omg.org/CORBA/Contained:1.0' then
    result := Pointer(IContained(self))
  else if repoid = 'IDL:omg.org/CORBA/IDLType:1.0' then
    result := Pointer(IIDLType(self))
  else
    result := inherited narrow_helper(repoid)
end;

procedure TValueDef.set_abstract_base_values(const val: IInterfaceList);
begin
  FAbstractBaseValues := val;
end;

procedure TValueDef.set_base_value(const val: IValueDef);
begin
  FBaseValue := val;
end;

procedure TValueDef.set_is_abstract(val: Boolean);
begin
  FAbstract := val
end;

procedure TValueDef.set_is_custom(val: Boolean);
begin
  FCustom := val;
end;

procedure TValueDef.set_is_truncatable(val: Boolean);
begin
  FTruncatable := val;
end;

procedure TValueDef.set_supported_interfaces(const val: IInterfaceList);
begin
  FSupportedInt := val
end;

constructor TValueDef.Create(const con: IContainer; id, name,
  version: string; is_custom, is_abstract: Boolean; base_value: IValueDef;
  is_truncatable: Boolean; abstract_base_values, supported_interfaces: IInterfaceList;
  initializers: PInitializerSeq);
begin
  inherited Create();
  FDefKind := dk_Value;
  FContainer := con;
  FName := name;
  FID := id;
  FVersion := version;
  FCustom := is_custom;
  FAbstract := is_abstract;
  FTruncatable := is_truncatable;
  FAbstractBaseValues := abstract_base_values;
  FSupportedInt := supported_interfaces;
  if initializers <> nil then
    FInitializers := initializers^;
end;

class function TValueDef._narrow(const obj: IORBObject): IValueDef;
var
  ptr: Pointer;
begin
  result := nil;
  if obj <> nil then
    begin
      ptr := obj.narrow_helper('IDL:omg.org/CORBA/ValueDef:1.0');
      if ptr <> nil then
        result := IValueDef(ptr)
      else
        ;//assert(false,'not imlplemented');
    end;
end;

function TValueDef.describe_value: FullValueDescription;
var
  i,j : integer;
  list : IInterfaceList;
  cont: IContained;
begin
  with result do
    begin
      name := FName;
      id := FID;
      is_abstract := FAbstract;
      is_custom := FCustom;
      cont := TContained._narrow(FContainer as IORBObject);
      if cont <> nil then
        defined_in := cont.id
      else
        defined_in := '';
      Version := FVersion;
      list := (self as IContainer).contents(dk_Operation);
      setLength(operations,list.count);
      for i := 0 to Pred(list.count) do
        operations[i] := (list[i] as IOperationDef)._describe;
      j := 0;
      setLength(attributes,j);
      assert(false,'not implemented');
    end;
end;


procedure TValueDef.check_overloaded(id,oname: string);
var
  i : integer;
begin
  for i := 0 to Pred(FContained.count) do
    with IContained(FContained[i]) do
      if (def_kind in [dk_Operation,dk_Attribute]) and  (name = oname) then
        dorb_throw('Overloading');
end;

//*********************************************************************
//  TValueMemberDef
//*********************************************************************
constructor TValueMemberDef.Create(const con: IContainer; id, name,
  version: string; mtype: IIDLType; acc: TVisibility);
begin
  inherited Create();
  FDefKind := dk_ValueMember;
  FContainer := con;
  FID := id;
  FName := name;
  FType := mtype;
  FVersion := version;
  FAccess := acc;
end;

function TValueMemberDef._type: ITypeCode;
begin
  result := FType._type
end;

function TValueMemberDef.get_access: TVisibility;
begin
  result := FAccess
end;

function TValueMemberDef.get_type_def: IIDLType;
begin
  result := FType;
end;

procedure TValueMemberDef.set_access(val: TVisibility);
begin
 FAccess := val;
end;

procedure TValueMemberDef.set_type_def(const val: IIDLType);
begin
  FType := val;
end;

function TValueMemberDef.narrow_helper(const repoid: string): Pointer;
begin
  if repoid = 'IDL:omg.org/CORBA/ValueMemberDef:1.0' then
    result := Pointer(IValueMemberDef(self))
  else
    result := inherited narrow_helper(repoid);
end;

class function TValueMemberDef._narrow(const obj: IORBObject): IValueMemberDef;
var
  ptr: Pointer;
begin
  result := nil;
  if obj <> nil then
    begin
      ptr := obj.narrow_helper('IDL:omg.org/CORBA/ValueMemberDef:1.0');
      if ptr <> nil then
        result := IValueMemberDef(ptr)
      else
        //'not imlplemented'
    end;
end;

end.
